#include <arpa/inet.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <poll.h>
#include <sys/time.h>
#include <termio.h>
#include <time.h>

#include <linux/serial.h>

#include "arm-probe.h"

struct aep aeps[MAX_PROBES];
struct pollfd pollfds[MAX_PROBES];
int count_pollfds = 0;
int verbose = 0;
int average_len = 1;
char channel_name[64];
int channel = 0;
int ms_holdoff = 0;
int ms_capture = 0;
int exit_after_capture = 0;
int awaiting_capture = 0;
char configuration_name[64] = "default_configuration";
int just_power = 0;
char stdinbuf[512];
char stdinline[512] = "";
int stdinpos = 0;
int stdinlen = 0;
int require_off = 0;
int show_raw = 0;
int decimate = 1;
int no_correction = 0;
int auto_zero = 0;
int adaptive = 0;

char config_filepath[256] = "./config";

#define TRIG_HYSTERESIS_MV 200
#define TRIG_HYSTERESIS_MW 200

/* squelch samples below this */
unsigned int mv_min = 400;
unsigned int mw_min = -1000000;
unsigned int mw_min_plus_hyst = 0;
unsigned int trigger_filter_us = 400;
unsigned int end_trigger_filter_us = 500000;
unsigned int ms_pretrigger = 0;

#define ADC_COUNTS_PER_VOLT_CH1 1000.0
#define ADC_COUNTS_PER_VOLT_CH2 10000.0

/* in V */

struct interp {
	double actual;
	double raw;
};

#define B 0.000258

struct interp fix_771mV[] = {
		{ 0.000000, 0.000258 - B },
		{ 0.000160, 0.000268 - B },
                { 0.000568, 0.000359 - B },
                { 0.000923, 0.000443 - B },
                { 0.001349, 0.000595 - B },
                { 0.001910, 0.000875 - B },
                { 0.002284, 0.001126 - B },
                { 0.003007, 0.001766 - B },
                { 0.003411, 0.002140 - B },
                { 0.003943, 0.002642 - B },
                { 0.004269, 0.002944 - B },
                { 0.004677, 0.003323 - B },
                { 0.005053, 0.003684 - B },
                { 0.005422, 0.004057 - B },
                { 0.006370, 0.005025 - B },
                { 0.007108, 0.005770 - B },
                { 0.008075, 0.006737 - B },
                { 0.008804, 0.007475 - B },
                { 0.009734, 0.008369 - B },
                { 0.010481, 0.009130 - B },
                { 0.012162, 0.010870 - B },
                { 0.014196, 0.012890 - B },
                { 0.016268, 0.014970 - B },
                { 0.018317, 0.017024 - B },
                { 0.020375, 0.019036 - B },
                { 0.024884, 0.023467 - B },
                { 0.028644, 0.027257 - B },
                { 0.033287, 0.032000 - B },
                { 0.037051, 0.035769 - B },
                { 0.041469, 0.040192 - B },
                { 0.049413, 0.048233 - B },
                { 0.057643, 0.056520 - B },
                { 0.061759, 0.060652 - B },
                { 0.072068, 0.071021 - B },
                { 0.082604, 0.081653 - B },
                { 0.103219, 0.102417 - B },
                { 0.124168, 0.123468 - B },
		{ 0.144894, 0.144408 - B },
		{ 0.165917, 0.161909 - B },
};

double interp(double in, struct interp *fix, int fix_count)
{
	int n;
	double factor;

	for (n = 0; n < fix_count; n++) {
		if (in > fix[n].raw)
			continue;
		if (n == 0)
			return in;
		
		factor = (in - fix[n - 1].raw) / (fix[n].raw - fix[n - 1].raw);

		in = fix[n - 1].actual + 
			(factor * (fix[n].actual - fix[n - 1].actual));

		return in;
	}
	return in;
}

double correct(double common_mode_V, double in)
{
	if (no_correction)
		return in;

//	fprintf(stderr, "%f\n", common_mode_V);

	return interp(in, fix_771mV, sizeof fix_771mV / sizeof fix_771mV[0]);
}

void pull_stdin_if_needed(void)
{
	int n;

	if (stdinpos != stdinlen)
		return;

	n = read(0, stdinbuf, sizeof stdinbuf);
	if (n < 0) {
		if (verbose)
			fprintf(stderr, "pull_stdin_if_needed: n=%d\n", n);
		return;
	}
	if (n == 0 && verbose)
		fprintf(stderr, "pull_stdin_if_needed n = 0\n");
	stdinpos = 0;
	stdinlen = n;
}

int get_stdin_line(char *buf, int maxlen)
{
	while (maxlen--) {
		if (stdinpos == stdinlen) {
			pull_stdin_if_needed();
			if (stdinpos == stdinlen)
				return -1;
		}
		*buf++ = stdinbuf[stdinpos++];
		if (buf[-1] == '\n') {
			buf[-1] = '\0';
			return 0;
		}
	}
	return -1;
}

static void init_aep(struct aep *aep, const char *device_filepath)
{
	int n, m;
	struct aep_channel *ch;

	aep->head = 0;
	aep->tail = 0;
	aep->predicted_frame = 0;
	aep->state = APP_INIT_MAGIC;
	aep->invalid = 0;
	aep->counter = 0;
	aep->samples = 0;
	aep->started = 0;
	aep->ignore = 0;
	aep->triggered = 0;
	aep->selected_channel = -1;

	strncpy(aep->dev_filepath, device_filepath,
						sizeof aep->dev_filepath - 1);
	aep->dev_filepath[sizeof aep->dev_filepath - 1] = '\0';

	for (n = 0; n < CHANNELS_PER_PROBE; n++) {
		ch = &aep->ch[n];	

		ch->voffset[0] = 0;
		ch->vnoise[0] = 0;
		ch->voffset[1] = 0;
		ch->vnoise[1] = 0;
		ch->rshunt = 0;
		sprintf(ch->channel_name, "%s-%d", device_filepath, n);
		ch->pretrig_ring = NULL;
		ch->ring_samples = 0;
		ch->trigger_filter = 0;
		ch->pretrigger_samples_taken = 0;
		avg_mean_us_init(&ch->avg_mean_voltage, average_len);
		avg_mean_us_init(&ch->avg_mean_current, average_len);
		ch->decimation_counter = 0;
		for (m = 0; m < sizeof(ch->min) / sizeof(ch->min[0]); m++) {
			ch->min[m] = 999;
			ch->max[m] = 0;
		}
		ch->samples_seen = 0;
	}
}


void add_pollfd(int fd, int events)
{
	pollfds[count_pollfds].fd = fd;
	pollfds[count_pollfds].events = events;
	pollfds[count_pollfds++].revents = 0;
}

void remove_pollfd(int fd)
{
	int n;

	for (n = 0; n < count_pollfds; n++)
		if (pollfds[n].fd == fd) {
			while (n < count_pollfds) {
				pollfds[n] = pollfds[n + 1];
				n++;
			}
			count_pollfds--;
		}
}


static void probe_fini(struct aep *aep)
{
	unsigned char c = AEPC_STOP;
	int n;

//	if (aep->started)
		write(aep->fd, &c, 1);

	for (n = 0; n < CHANNELS_PER_PROBE; n++) {
		if (aep->ch[n].pretrig_ring)
			free(aep->ch[n].pretrig_ring);

		avg_mean_us_free(&aep->ch[n].avg_mean_voltage);
		avg_mean_us_free(&aep->ch[n].avg_mean_current);
	}

	if (verbose)
		fprintf(stderr, "closing %d\n", aep->fd);
	close(aep->fd);
}

#define IGNORE_NEGATIVE_PRIOR_TO_SAMPLES 2000
#define NEGATIVE_ADJUST_SCALING_FACTOR 0.01

void track_limits_convert_to_current(struct aep_channel *ch, double *v1, double *v2)
{
	double pwr;

	if ((*v1 - ch->voffset[0]) < ch->min[3])
		ch->min[3] = *v1 - ch->voffset[0];

	if ((*v1 - ch->voffset[0]) > ch->max[3])
		ch->max[3] = *v1 - ch->voffset[0];

	if ((*v2 - ch->voffset[1]) < ch->min[4])
		ch->min[4] = *v2 - ch->voffset[1];

	if ((*v2 - ch->voffset[1]) > ch->max[4])
		ch->max[4] = *v2 - ch->voffset[1];

	if (adaptive && !auto_zero) {
		if (*v1 < 0) {
			if (ch->samples_seen > IGNORE_NEGATIVE_PRIOR_TO_SAMPLES) {
				if (-(*v1) > ch->vnoise[0] / 2) {
					if (verbose)
						fprintf(stderr, "*** reducing voltage baseline by %fV to %fV\n", *v1, ch->voffset[0] - *v1);
					ch->voffset[0] -= *v1 * NEGATIVE_ADJUST_SCALING_FACTOR;
				}
			}
			*v1 = 0;
		}
		if (*v2 < 0) {
			if (ch->samples_seen > IGNORE_NEGATIVE_PRIOR_TO_SAMPLES) {
				if (-(*v2) > ch->vnoise[1] / 2) {
					if (verbose)
						fprintf(stderr, "*** reducing shunt voltage baseline by %fV to %fV\n", *v2, ch->voffset[1] - *v2);
					ch->voffset[1] -= *v2 * NEGATIVE_ADJUST_SCALING_FACTOR;
				}
			}
			*v2 = 0;
		}
	}

	*v2 = correct(*v1, *v2) / ch->rshunt;

	pwr = *v1 * *v2;
	if (*v1 > ch->max[0])
		ch->max[0] = *v1;
	if (*v1 < ch->min[0])
		ch->min[0] = *v1;
	if (*v2 > ch->max[1])
		ch->max[1] = *v2;
	if (*v2 < ch->min[1])
		ch->min[1] = *v2;
	if (pwr > ch->max[2])
		ch->max[2] = pwr;
	if (pwr < ch->min[2])
		ch->min[2] = pwr;
}

int process_sample(struct aep *aep)
{
	struct aep_channel *ch = &aep->ch[aep->selected_channel];
	double v1, v2;

	if (no_correction) {
		ch->voffset[0] = 0;
		ch->voffset[1] = 0;
	}

	if (ch->pretrig_ring) {
		ch->pretrig_ring[ch->head].v = aep->voltage;
		ch->pretrig_ring[ch->head].i = aep->current;
		ch->head++;
		if (ch->head == ch->ring_samples)
			ch->head = 0;
		if (ch->head == ch->tail)
			ch->tail++;
		if (ch->tail == ch->ring_samples)
			ch->tail = 0;

		if (ch->pretrigger_samples_taken < ch->ring_samples)
			ch->pretrigger_samples_taken++;
	}

	if (aep->triggered && aep->voltage < mv_min && aep->samples > 1000) {

		ch->trigger_filter++;
		if (ch->trigger_filter < end_trigger_filter_us / 100)
			goto unripe;

		fprintf(stderr, "trigger session completed "
				"%dmV vs %dmV\n", aep->voltage, mv_min);
		avg_mean_us_flush(&ch->avg_mean_voltage);
		avg_mean_us_flush(&ch->avg_mean_current);
		aep->ignore = 0;
		aep->triggered = 0;
		ch->trigger_filter = 0;
		ch->pretrigger_samples_taken = 0;
		ch->head = 0;
		ch->tail = 0;
		if (require_off) {
			fprintf(stderr, "marking capture for "
					"device %d done\n", aep->index);
			awaiting_capture &= ~(1 << aep->index);
		}
		return 0;
	} else
		if (aep->triggered)
			ch->trigger_filter = 0;
unripe:

	ch->samples_seen++;

	if (!aep->triggered && (awaiting_capture & (1 << aep->index))) {
		avg_mean_us_add(&ch->avg_mean_voltage, aep->voltage);
		avg_mean_us_add(&ch->avg_mean_current, aep->current);

		v1 = (avg_mean_us(&ch->avg_mean_voltage) / ADC_COUNTS_PER_VOLT_CH1) + ch->voffset[0];
		v2 = (avg_mean_us(&ch->avg_mean_current) / ADC_COUNTS_PER_VOLT_CH2) + ch->voffset[1];

		track_limits_convert_to_current(ch, &v1, &v2);

		if (((aep->voltage > mv_min + TRIG_HYSTERESIS_MV) || !mv_min) &&
			(((v1 * v2) > (mw_min_plus_hyst) / 1000.0) || !mw_min)) {

			ch->trigger_filter++;
			if (ch->trigger_filter > trigger_filter_us / 100) {
							
				aep->triggered = 1;
				ch->trigger_filter = 0;
				fprintf(stderr, "%s: triggered "
						"%.2fV %.5fW pretrig %d "
						"(%dms) selected ch = %d\n",
						ch->channel_name, v1,
						v1 * v2,
						ch->ring_samples,
						ch->ring_samples / 10,
						aep->selected_channel);
			}
		} else
			ch->trigger_filter = 0;

		aep->rx_count++;
		if ((aep->rx_count & 0x3ff) == 0) {
			fprintf(stderr, "%s: pretrigger %.2fV "
					"%.4fA (ofs-only:%fV corr:%fV) %fW",
					ch->channel_name, v1, v2,
					avg_mean_us(&ch->avg_mean_current) / ADC_COUNTS_PER_VOLT_CH2 + ch->voffset[1],
					correct(v1, avg_mean_us(&ch->avg_mean_current) / ADC_COUNTS_PER_VOLT_CH2 + ch->voffset[1]),
					v1 * v2);

			if (ch->pretrigger_samples_taken < ch->ring_samples)
				fprintf(stderr, " (filling pretrig)");

			fprintf(stderr, "                      \r");
		}
	}

	if (aep->ignore || !aep->triggered)
		return 0;

	if (!aep->samples && ms_holdoff)
		fprintf(stderr, "%s: post-trigger holdoff for"
				     " %dms\n", ch->channel_name, ms_holdoff);

	aep->samples++;

	if (aep->samples / 10 < ms_holdoff)
		return 0;
	if (aep->samples - 1 == ms_holdoff * 10)
		fprintf(stderr, "%s: Starting capture\n", ch->channel_name);

	if (ch->pretrig_ring) {
		avg_mean_us_add(&ch->avg_mean_voltage,
						ch->pretrig_ring[ch->tail].v);
		avg_mean_us_add(&ch->avg_mean_current,
						ch->pretrig_ring[ch->tail].i);
		ch->tail++;
		if (ch->tail == ch->ring_samples)
			ch->tail = 0;
	} else {
		avg_mean_us_add(&ch->avg_mean_voltage, aep->voltage);
		avg_mean_us_add(&ch->avg_mean_current, aep->current);
	}
	v1 = (avg_mean_us(&ch->avg_mean_voltage) / ADC_COUNTS_PER_VOLT_CH1) + ch->voffset[0];
	v2 = (avg_mean_us(&ch->avg_mean_current) / ADC_COUNTS_PER_VOLT_CH2) + ch->voffset[1];

	track_limits_convert_to_current(ch, &v1, &v2);

	if (auto_zero && v1 > 1.0) {
		fprintf(stderr, "**** Autozero seeing >1V... probe should not be in powered target for autozero! ****\n");
		return -2;
	}

	ch->decimation_counter++;
	if (ch->decimation_counter != decimate)
		goto done;
	ch->decimation_counter = 0;

	if (auto_zero) {
		fprintf(stderr, "Autozero for %s done         \n", ch->channel_name);
		ch->voffset[0] = -v1;
		ch->vnoise[0] = ch->max[3] - ch->min[3];
		ch->voffset[1] = -avg_mean_us(&ch->avg_mean_current) / ADC_COUNTS_PER_VOLT_CH2;
		ch->vnoise[1] = ch->max[4] - ch->min[4];
		if (configure(aep, aep->dev_filepath, config_filepath, ch) < 0)
			fprintf(stderr, "Failed to write autozero info to config\n");
		goto done;
	}

	if (stdinline[0]) {
		if (just_power) {
			printf("%s %.5f\n", stdinline, v1 * v2);
		} else
			printf("%s %.2f %.4f %.5f\n", stdinline,
						v1, v2, v1 * v2);
		if (get_stdin_line(stdinline, sizeof stdinline))
			stdinline[0] = '\0';
		goto done;
	}

	if (just_power)
		printf("%lu %.5f\n", aep->samples, v1 * v2);
	else
		printf("%lu %.2f %.4f %.5f\n",
					aep->samples, v1, v2, v1 * v2);
done:

	if ((aep->samples & 0x7ff) == 0) {
		if (show_raw)
			fprintf(stderr, "%s: %.2fV %fA (shunt: avg+ofs:%fV corrected:%fV offset:%fV ADC:%d) %fW ",
				ch->channel_name, v1, v2, (avg_mean_us(
				&ch->avg_mean_current) / ADC_COUNTS_PER_VOLT_CH2) +
				ch->voffset[1], correct(v1, avg_mean_us(
				&ch->avg_mean_current) / ADC_COUNTS_PER_VOLT_CH2 +
				ch->voffset[1]), ch->voffset[1], aep->current,
				v1 * v2);
		else 
			fprintf(stderr, "%s: %.2fV %fA %fW ",
				ch->channel_name, v1, v2, v1 * v2);    

		if (ms_capture) {
			int u, t;
			fprintf(stderr, " [");
			t = (aep->samples * 20) / (ms_capture * 10);
			for (u = 0; u < t; u++)
				fprintf(stderr, "*");
			while (u++ < 20)
				fprintf(stderr, "-");
			fprintf(stderr, "]");
		}

		if (ch->pretrigger_samples_taken < ch->ring_samples)
			fprintf(stderr, "  (filling pretrig)");
                                                                                
		fprintf(stderr, "                           \r");
	}

	if (ms_capture && (aep->samples / 10 - ms_holdoff) > ms_capture) {
		fprintf(stderr, "%s: %dms capture complete\n",
				    ch->channel_name, ms_capture);
		aep->ignore = 1;
		if (!require_off)
			awaiting_capture &= ~(1 << aep->index);
		else
			fprintf(stderr, "(waiting for OFF)\n");
	}

	return 0;
}


int service_aep(struct aep *aep, int samples)
{
	unsigned char v;
	static unsigned char cmd[3] = { AEPC_CONFIG, 0, 0x80 | 6 };
	int bad = 0;

	if (aep->selected_channel >= 0)
		cmd[1] = aep->selected_channel;

	while (aep->head != aep->tail && samples--) {

		v = aep->ring[aep->head++];
		if (aep->head == sizeof(aep->ring))
			aep->head = 0;

		if (verbose)
			fprintf(stderr, "%02X ", v);

		switch (aep->state) {

		case APP_INIT_MAGIC:
			switch (aep->counter++) {
			case 0:
				if (v != 0xac)
					return -1;
				break;
			case 1:
			case 2:
			case 3:
			case 4:
			case 5:
			case 6:
			case 7:
			case 8:
				if (v != 0xff)
					return -1;
				break;
			case 63:
				if (verbose)
					fprintf(stderr,
					     "magic ok, asking for version\n");
				v = AEPC_VERSION;
				if (write(aep->fd, &v, 1) != 1)
					return -1;
				aep->state++;
				aep->counter = 0;
				break;
			default:
				break;
			}
			break;

		case APP_INIT_VERSION:
			switch (aep->counter++) {
			case 0:
				if (v != 0xac)
					return -1;
				break;
			case 1:
				aep->version = v;
				break;
			case 2:
				aep->version |= v << 8;
				break;
			case 3:
				aep->version |= v << 16;
				break;
			case 4:
				aep->version |= v << 24;
				break;
			case 63:
				if (verbose)
					fprintf(stderr,
					    "version ok, asking for vendor\n");
				v = AEPC_VENDOR;
				if (write(aep->fd, &v, 1) != 1)
					return -1;
	        		aep->state++;
				aep->counter = 0;
				break;
			default:
				break;
			}
			break;

		case APP_INIT_VENDOR:
			switch (aep->counter++) {
			case 0:
				if (v != 0xac)
					return -1;
				break;
			case 63:
				if (verbose)
					fprintf(stderr,
					       "vendor ok, asking for rate\n");
				v = AEPC_RATE;
				if (write(aep->fd, &v, 1) != 1)
					return -1;
	        		aep->state++;
				aep->counter = 0;
				aep->name[63] = '\0';
				break;
			default:
				aep->name[aep->counter - 2] = v;
				break;
			}
			break;


		case APP_INIT_RATE:
			switch (aep->counter++) {
			case 0:
				if (v != 0xac)
					return -1;
				break;
			case 1:
				aep->rate = v;
				break;
			case 2:
				aep->rate |= v << 8;
				break;
			case 63:
				if (verbose)
					fprintf(stderr,
						    "rate ok, doing config\n");
	        		aep->state++;
				aep->counter = 0;
				if (write(aep->fd, &cmd[0], 3) != 3)
					return -1;
				break;
			default:
				break;
			}
			break;

		case APP_INIT_CONFIG:
			
			switch (aep->counter++) {
			case 0:
				if (v != 0xac)
					return -1;
				break;
			case 63:
				v = AEPC_START;
				if (write(aep->fd, &v, 1) != 1)
					return -1;

				fprintf(stderr, "Started new probe instance "
						"%s %s %08X %d\n",
						aep->dev_filepath, aep->name,
						      aep->version, aep->rate);

				aep->done_config |= 2;
				aep->started = 1;
				aep->state = APP_FRAME_L;
				aep->counter = 0;
				break;
			default:
				break;
			}
			break;

		case APP_FRAME_L:
			if (v == (aep->predicted_frame & 0xff)) {
				aep->state++;
				break;
			}
			/* lost sync... */
			aep->invalid = 1;
			aep->state++;	
			break;
		case APP_FRAME_H:

			if (!aep->invalid && v == aep->predicted_frame >> 8) {
				aep->state++;
				aep->predicted_frame++;
				break;
			}
			/* lost sync - slip one byte */
			aep->predicted_frame = v;
			aep->invalid = 1;
			aep->state++;
			break;
		case APP_VOLTAGE_L:
			if (aep->invalid) {
				aep->predicted_frame |= v << 8;
				aep->predicted_frame++;
				aep->state++;
				break;
			}
			aep->voltage = v;
			aep->state++;
			break;

		case APP_VOLTAGE_H:
			aep->voltage |= v << 8;
			aep->state++;
			break;

		case APP_CURRENT_L:
			aep->current = v;
			aep->state++;
			break;

		case APP_CURRENT_H:
			aep->current |= v << 8;
			if (aep->invalid) {
				if (verbose)
					fprintf(stderr, "%s: *\n",
							aep->dev_filepath);
				bad++;
				if (bad > 8) {
					fprintf(stderr, "too much lost sync\n");
					return -1;
				}
				aep->state++;
				break;
			}
			aep->state = APP_FRAME_L;

			if (aep->selected_channel == -1) {
				//fprintf(stderr, "ignoring %s\n", aep->dev_filepath);
				break;
			}
			if (aep->current > 1620)
				fprintf(stderr, "*** Shunt current near 165mV limit: clipping ***\n");

			return process_sample(aep);

		case APP_SWALLOW:
			aep->invalid = 0;
			aep->state = APP_FRAME_L;
			break;

		}
	}

	return 0;
}


static struct option options[] = {
	{ "help",	no_argument,		NULL, 'h' },
	{ "verbose",	no_argument,		NULL, 'v' },
	{ "raw",	no_argument,		NULL, 'r' }, 
	{ "mean",	required_argument,	NULL, 'm' },
	{ "decimate",	required_argument,	NULL, 'd' },
	{ "mvtrigger",	required_argument,	NULL, 'q' },
	{ "mwtrigger",	required_argument,	NULL, 'w' },
	{ "channel",	required_argument,	NULL, 'c' },
	{ "mstrighold",	required_argument, 	NULL, 't' },
	{ "ustrigfilter", required_argument,	NULL, 'f' },
	{ "mslength",	required_argument, 	NULL, 'l' },
	{ "exitoncap",	no_argument,		NULL, 'x' },
	{ "justpower",	no_argument,		NULL, 'j' },
	{ "config",	required_argument,	NULL, 'C' },
	{ "offexit",	no_argument,		NULL, 'o' },
	{ "mspretrigger", required_argument,	NULL, 'p' },
	{ "nocorrection", no_argument,		NULL, 'n' },
	{ "autozero",	no_argument,		NULL, 'z' },
	{ "adaptive",	no_argument,		NULL, 'a' },
	{ NULL, 0, 0, 0 }
};

int main(int argc, char *argv[])
{
	static unsigned char zero[8] = { 0, 0, 0, 0, 0, 0, 0, 0, };
	time_t t;
	struct tm *tmp;
	char hostname[200];
	char date[200];
	int fd, loop = 1, n = 0;
	struct termios tty;
	struct serial_struct sinfo;
	struct aep_channel *sel_ch = NULL;
	struct aep *sel_aep = NULL;
	unsigned char *p;
	unsigned char c;
	char filepath[20];
	struct timeval tv;
	int highest = 0;
	int scan = 0;
	int scans = 0;
	int budget;
	int len;
	int m;

	void sighandler(int sig)
	{
		loop = 0;
	}

	signal(SIGINT, sighandler);

	while (n >= 0) {
		n = getopt_long(argc, argv, "xl:t:c:q:hvm:jC:w:f:op:rd:nza",
							options, NULL);
		if (n < 0)
			continue;
		switch (n) {
		case 'a':
			adaptive = 1;
			break;
		case 'z':
			fprintf(stderr, "\n\n *** for autozero, you must short both sides of channel sense leads to 0V ***\n\n");
			no_correction = 1;
			mv_min = 0;
			mw_min = 0;
			auto_zero = 1;
			average_len = 50000;
			ms_capture = average_len / 10;
			decimate = average_len;
			exit_after_capture = 1;
			break;
		case 'n':
			no_correction = 1;
			break;
		case 'd':
			decimate = atoi(optarg);
			break;
		case 'r':
			show_raw = 1;
			break;
		case 'p':
			ms_pretrigger = atoi(optarg);
			break;
		case 'o':
			require_off = 1;
			break;
		case 'f':
			trigger_filter_us = atoi(optarg);
			break;
		case 'w':
			mw_min = atoi(optarg);
			if (mw_min)
				mw_min_plus_hyst = mw_min + (mw_min / 10);
			break;
		case 'j':
			just_power = 1;
			break;
		case 'C':
			strncpy(config_filepath, optarg,
						sizeof config_filepath - 1);
			config_filepath[sizeof config_filepath - 1] = '\0';
			break;
		case 't':
			ms_holdoff = atoi(optarg);
			break;
		case 'l':
			ms_capture = atoi(optarg);
			break;
		case 'c':
			strncpy(channel_name, optarg, sizeof channel_name - 1);
			channel_name[sizeof channel_name - 1] = '\0';
			break;
		case 'q':
			mv_min = atoi(optarg);
			break;
		case 'm':
			average_len = atoi(optarg);
			break;
		case 'v':
			verbose++;
			break;
		case 'x':
			exit_after_capture = 1;
			break;
		default:
		case 'h':
			fprintf(stderr, "Usage: arm-probe "
				"[--verbose -v] Increase debug on stderr\n"
				"[--mean -m <averagng depth>]\n"
				"[--mvsquelch -q <min mv>] suppress output "
					"below this voltage (defaut 400mV)\n"
				"[--channel -c <chan 0-2>] select channel "
					"(default 0)\n"
				"[--triggerms -t <ms>] delay in ms after "
					"power seen before capture\n"
				"[--length -l <ms>] period to capture after "
					"trigger (+holdoff if any)\n"
				"[--exitoncap -x] after requested capture "
					"(-l) completes on all channels, "
					"exit app\n"
			);
			exit(1);
		}
	}

	/* set stdin to noblock */

	if (isatty(0))
		goto post_si;

	m = 1;
	n = 0;
	while (m) {
		n++;
		if (n > 1000) {
			fprintf(stderr, "problem with stdin format\n");
			return 1;
		}
		if (get_stdin_line(stdinline, sizeof stdinline)) {
			fprintf(stderr, "get_stdin_line failed\n");
			m = 0;
			continue;
		}
		if (stdinline[0] != '#') {
			if (verbose)
				fprintf(stderr, "seen '%s'\n", stdinline);
			puts("#");
			m = 0;
			continue;
		}
		puts(stdinline);
	}
post_si:
	configure(NULL, "xx", config_filepath, NULL);
	
	printf("# configuration: %s\n", config_filepath);
	printf("# config_name: %s\n", configuration_name);
	printf("# trigger: %fV (hyst %fV) %fW (hyst %fW) %uus\n",
		mv_min / 1000.0, TRIG_HYSTERESIS_MV / 1000.0, 
		mw_min / 1000.0, TRIG_HYSTERESIS_MV / 1000.0,
		trigger_filter_us);

	fprintf(stderr, "Configuration: %s\n", configuration_name);

	t = time(NULL);
	tmp = localtime(&t);
	if (tmp) {
		if (strftime(date, sizeof date, "%a, %d %b %Y %T %z", tmp))
			printf("# date: %s\n", date);
	}

	if (!gethostname(hostname, sizeof hostname))
		printf("# host: %s\n", hostname);

	printf("#\n");

	while (loop) {

		gettimeofday(&tv, NULL);

		/* look for any new devices appearing */

		if (aeps[scan].fd > 0)
			goto bail;

		sprintf(filepath, "/dev/ttyACM%d", scan);
		fd = open(filepath, O_RDWR | O_NONBLOCK | O_EXCL | O_NOCTTY);

		if (fd < 0)
			goto bail;

		if (ioctl(fd, TIOCEXCL) < 0) {
			fprintf(stderr, "Wasn't able to open %s exclusive",
					filepath);
			close(fd);
			goto bail;
		}

		tcflush(fd, TCIOFLUSH);

		if (ioctl(fd, TIOCGSERIAL, &sinfo) == 0) {
			sinfo.closing_wait = ASYNC_CLOSING_WAIT_NONE;
			ioctl(fd, TIOCSSERIAL, &sinfo);
		}

		if (verbose)
			fprintf(stderr, "initing %s fd=%d\n", filepath, fd);

		/* enforce suitable tty state */

		memset (&tty, 0, sizeof tty);
		if (tcgetattr (fd, &tty)) {
			fprintf(stderr, "tcgetattr failed on %s\n", filepath);
			close(fd);
			goto bail;
		}

		tty.c_lflag &= ~(ISIG | ICANON | IEXTEN | ECHO | XCASE |
				ECHOE | ECHOK | ECHONL | ECHOCTL | ECHOKE);
		tty.c_iflag &= ~(INLCR | IGNBRK | IGNPAR | IGNCR | ICRNL |
				   IMAXBEL | IXON | IXOFF | IXANY | IUCLC);
		tty.c_oflag &= ~(ONLCR | OPOST | OLCUC | OCRNL | ONLRET);
		tty.c_cc[VMIN]  = 1;
		tty.c_cc[VTIME] = 0;
		tty.c_cc[VEOF] = 1;
		tty.c_cflag &=  ~(CBAUD | CSIZE | CSTOPB | PARENB | CRTSCTS);
		tty.c_cflag |= (8 | CREAD | 0 | 0 | 1 | CLOCAL);

		cfsetispeed(&tty, B115200);
		cfsetospeed(&tty, B115200);
		tcsetattr(fd, TCSANOW, &tty);

		init_aep(&aeps[scan], filepath);

		if (configure(&aeps[scan], filepath, config_filepath, NULL) < 0) {
			fprintf(stderr, "config for %s failed\n", filepath);
			close(fd);
			goto bail;
		}

		if (write(fd, zero, sizeof(zero)) != sizeof(zero)) {
			close(fd);
			goto bail;
		}
		if (verbose)
			fprintf(stderr, "sending reset\n");
		c = AEPC_RESET;
		if (write(fd, &c, 1) != 1) {
			close(fd);
			goto bail;
		}

		if (verbose)
			fprintf(stderr, "done reset\n");

		aeps[scan].fd = fd;
		aeps[scan].sec_last_traffic = tv.tv_sec;
		aeps[scan].index = scan;
	
		if (scan > highest)
			highest = scan;
		add_pollfd(fd, POLLIN | POLLERR);


		for (n = 0; n < CHANNELS_PER_PROBE; n++) {
			struct aep_channel *ch = &aeps[scan].ch[n];

			if (strcmp(ch->channel_name, channel_name))
				continue;
			aeps[scan].selected_channel = n;
			sel_ch = ch;
			sel_aep = &aeps[scan];
			if (!(awaiting_capture & (1 << scan))) {
				fprintf(stderr, "seen selected channel %s\n",
					ch->channel_name_pretty);
				if (stdinline[0]) {
					if (just_power)
						printf("%s %s\n", stdinline,
							ch->channel_name_pretty);
					else
						printf("%s %s(V) %s(A) %s(W)\n",
							stdinline,
							ch->channel_name_pretty,
							ch->channel_name_pretty,
							ch->channel_name_pretty);
					get_stdin_line(stdinline,
							sizeof stdinline);
				} else
					if (just_power)
						printf("sample %s(W)\n",
							ch->channel_name_pretty);
					else
						printf("sample %s(V) %s(A) %s(W)\n",
							ch->channel_name_pretty,
							ch->channel_name_pretty,
							ch->channel_name_pretty);
			}
			awaiting_capture |= 1 << scan;
		}

		aeps[scan].done_config |= 1;

bail:
		scans++;
		scan++;
		if (scan == MAX_PROBES)
			scan = 0;

		/* timeout will only matter when no probes connected */

		if (verbose)
			fprintf(stderr, "count_pollfds=%d, [0].events=%d, ",
					     count_pollfds, pollfds[0].events);
		n = poll(pollfds, count_pollfds, 100);
		if (verbose)
			fprintf(stderr, "n = %d %d\n", n, pollfds[0].fd);
		if (n < 0) {
			fprintf(stderr, "poll failed %d\n", count_pollfds);
			loop = 0;
			continue;
 		}

		/* somebody had something for us */

		for (m = 0; m <= highest; m++) {

			if (aeps[m].fd < 1)
				continue;

			/* check for timeout */

			if (tv.tv_sec > aeps[m].sec_last_traffic + 2) {
				fprintf(stderr, "%s: timeout\n",
						aeps[m].dev_filepath);
				goto died;
			}

			for (n = 0; n < count_pollfds; n++)
				if (aeps[m].fd == pollfds[n].fd)
					break;
			if (n == count_pollfds)
				continue;

			if (!pollfds[n].revents)
				continue;

			if (verbose)
				fprintf(stderr,"stuff for %d\n", m);

			aeps[m].sec_last_traffic = tv.tv_sec;

			/* work out how much of the ring we can fill */

			if (aeps[m].head <= aeps[m].tail) {
				budget = sizeof(aeps[m].ring) - aeps[m].tail;
				p = aeps[m].ring + aeps[m].tail;
			} else {
				budget = aeps[m].head - aeps[m].tail - 1;
				p = aeps[m].ring + aeps[m].tail;
			}

			if (!budget)
				continue;

			len = read(aeps[m].fd, p, budget);
			if (len <= 0) {
				fprintf(stderr, "failed to read data\n");
				goto died;
			}
			aeps[m].tail += len;

			if (verbose)
				fprintf(stderr, "%d fetched %d\n", m, len);

			if (aeps[m].tail == sizeof(aeps[m].ring))
				aeps[m].tail = 0;

			continue;

died:
			fprintf(stderr, "removing probe due to error...\n");
			if (aeps[m].selected_channel >= 0)
				awaiting_capture &= ~(1 << m);
			probe_fini(&aeps[m]);
			remove_pollfd(aeps[m].fd);
			aeps[m].fd = 0;
			m = highest;
		}


		/* 
		 * service the existing devices
		 * need to limit the amount of time spent in the service
		 * otherwise other probes will drop data
		 */

		for (n = 0; n <= highest; n++)
			if (aeps[n].fd > 0) {
				m = service_aep(&aeps[n], 512);
				if (m < 0) {
					if (m < -1) {
						loop = 0;
						continue;
					}
					if (verbose)
						fprintf(stderr,
							"service failed\n");
					probe_fini(&aeps[n]);
					remove_pollfd(aeps[n].fd);
					aeps[n].fd = 0;
				}
			}

		if (scans > (5 * MAX_PROBES) &&
				awaiting_capture == 0 && exit_after_capture) {
			fprintf(stderr, "done all capture\n");
			loop = 0;
		}

	}

	if (adaptive && sel_aep)
		if (configure(sel_aep, sel_aep->dev_filepath,
						  config_filepath, sel_ch) < 0)
			fprintf(stderr, "adaptive: failed to update config\n");

	if (sel_ch && sel_ch->min[0] != 999)
		fprintf(stderr, "  Voltage: %.2fV min,     %.2fV max          \n"
				"  Current: %.4fA min,   %.4fA max\n"
				"  Power:   %fW min, %fW max\n",
					sel_ch->min[0], sel_ch->max[0],
					sel_ch->min[1], sel_ch->max[1],
					sel_ch->min[2], sel_ch->max[2]);

	for (n = 0; n < highest; n++)
		if (aeps[n].fd > 0)
			probe_fini(&aeps[n]);

	fprintf(stderr, "exited\n");
	
	return 0;
}

